/* $%BEGINLICENSE%$
 Copyright (c) 2012, Oracle and/or its affiliates. All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; version 2 of the
 License.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 02110-1301  USA

 $%ENDLICENSE%$ */
 

#include "admin-plugin-stats.h"
#include "network-mysqld-lua.h"

#define C(x) x, sizeof(x) -1
#define S(x) x->str, x->len

/**
 * resultset_cleanup:
 * @fields: array of MYSQL_FIELD
 * @rows: array of row 
 * @row: single row of values for MYSQL_FIELD
 *
 * frees the fields, rows and each single row of a resulset
 */
static void resultset_cleanup(GPtrArray *fields, GPtrArray *rows, GPtrArray *row) {

	gsize i, j;

	if (fields) {
		network_mysqld_proto_fielddefs_free(fields);
		fields = NULL;
	}

	if (rows) {
		for (i = 0; i < rows->len; i++) {
			row = rows->pdata[i];

			for (j = 0; j < row->len; j++) {
				g_free(row->pdata[j]);
			}

			g_ptr_array_free(row, TRUE);
		}
		g_ptr_array_free(rows, TRUE);
		rows = NULL;
	}
}

/**
 * resultset_send:
 * @descriptor: descriptor for MYSQL_FIELD
 * @value: value for descriptor
 * @con: a #network_mysqld_con
 *
 * send a resultset back to the client
 */
static void resultset_send(gchar *descriptor, long int value, network_mysqld_con *con) {

	GPtrArray *fields, *rows, *row;
	MYSQL_FIELD *field;

	fields = NULL;
	rows = NULL;
	row = NULL;

	fields = network_mysqld_proto_fielddefs_new();

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup(descriptor);
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	rows = g_ptr_array_new();
	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("%ld", value));
	g_ptr_array_add(rows, row);

	network_mysqld_con_send_resultset(con->client, fields, rows);

	resultset_cleanup(fields, rows, row);
}

/**
 * resultset_helper:
 * @con: a #network_mysqld_con
 *
 * send the helper resultset to the client
 */
static void resultset_helper(network_mysqld_con *con) {

	GPtrArray *fields, *rows, *row;
	MYSQL_FIELD *field;

	fields = NULL;
	rows = NULL;
	row = NULL;

	fields = network_mysqld_proto_fielddefs_new();

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("command");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("description");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	rows = g_ptr_array_new();

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW help"));
	g_ptr_array_add(row, g_strdup_printf("show this help"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW AVAILABLE STATS"));
	g_ptr_array_add(row, g_strdup_printf("lists all the available statistics"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW BACKENDS STATS"));
	g_ptr_array_add(row, g_strdup_printf("lists all backends and their statistics"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW <plugin> STATUS"));
	g_ptr_array_add(row, g_strdup_printf("list all gathered statistics for <plugin>"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW <plugin> <stat>"));
	g_ptr_array_add(row, g_strdup_printf("list <stat> gathered statistic for <plugin>"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW <max|current> CONNECTIONS"));
	g_ptr_array_add(row, g_strdup_printf("list gathered max, or avg latency statistic for MySQL Proxy"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW <min|max|avg> LATENCY"));
	g_ptr_array_add(row, g_strdup_printf("list gathered min, max or avg latency statistic for MySQL Proxy"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW STATUS"));
	g_ptr_array_add(row, g_strdup_printf("list all global gathered statistics for MySQL Proxy"));

	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("SHOW <stat>"));
	g_ptr_array_add(row, g_strdup_printf("list <stat> gathered statistic for MySQL Proxy"));

	g_ptr_array_add(rows, row);

	network_mysqld_con_send_resultset(con->client, fields, rows);

	resultset_cleanup(fields, rows, row);	
}

/*
 * DOC-COMMENT THIS!
 */
static void resultset_stats_available(network_mysqld_con *con) {

	GPtrArray *fields, *rows, *row;
	MYSQL_FIELD *field;

	fields = NULL;
	rows = NULL;
	row = NULL;

	fields = network_mysqld_proto_fielddefs_new();

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("statistics");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	rows = g_ptr_array_new();

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("uptime"));
	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("latency"));
	g_ptr_array_add(rows, row);

	row = g_ptr_array_new();
	g_ptr_array_add(row, g_strdup_printf("connections"));
	g_ptr_array_add(rows, row);

	network_mysqld_con_send_resultset(con->client, fields, rows);

	resultset_cleanup(fields, rows, row);
}

/*
 * DOC-COMMENT THIS!
 */
static void resultset_stats_backends(network_mysqld_con *con) {

	int i;

	GPtrArray *fields, *rows, *row;
	MYSQL_FIELD *field;

	fields = NULL;
	rows = NULL;
	row = NULL;

	fields = network_mysqld_proto_fielddefs_new();

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("backend_ndx");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("address");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("state");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("type");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("connected_clients");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("uuid");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	rows = g_ptr_array_new();

	g_mutex_lock(backends_stats->backends_mutex);

	for (i = 0; i < backends_stats->backends->len; i++) {

		network_backend_t *backend = backends_stats->backends->pdata[i];

		row = g_ptr_array_new();
		g_ptr_array_add(row, g_strdup_printf("%d", i));
		g_ptr_array_add(row, g_strdup_printf("%s", backend->addr->name->str));
		
		switch(backend->state) {

			case 0:
				g_ptr_array_add(row, g_strdup_printf("unknown"));
				break;
			case 1:
				g_ptr_array_add(row, g_strdup_printf("up"));
				break;
			case 2:
				g_ptr_array_add(row, g_strdup_printf("down"));
				break;
		}

		switch(backend->type) {

			case 0:
				g_ptr_array_add(row, g_strdup_printf("unknown"));
				break;
			case 1:
				g_ptr_array_add(row, g_strdup_printf("rw"));
				break;
			case 2:
				g_ptr_array_add(row, g_strdup_printf("ro"));
				break;
		}
		
		g_ptr_array_add(row, g_strdup_printf("%d", backend->connected_clients));

		if(backend->uuid == NULL)
			g_ptr_array_add(row, g_strdup_printf("null"));
		else
			g_ptr_array_add(row, g_strdup_printf("%s", backend->uuid->str));

		g_ptr_array_add(rows, row);
	}

	g_mutex_unlock(backends_stats->backends_mutex);

	network_mysqld_con_send_resultset(con->client, fields, rows);

	resultset_cleanup(fields, rows, row);
}

/**
 * resultset_stats_all_plugin:
 * @con: a #network_mysqld_con
 * @list: GSList of #plugin_stats_aux
 *
 * send all the gathered statistics of a specific plugin to the client
 */
static void resultset_stats_all_plugin(network_mysqld_con *con, GSList *list) {

	GPtrArray *fields, *rows, *row;
	MYSQL_FIELD *field;

	GSList *iterator = NULL;

	fields = NULL;
	rows = NULL;
	row = NULL;

	fields = network_mysqld_proto_fielddefs_new();

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("statistics");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("value");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	rows = g_ptr_array_new();

	for(iterator = list; iterator; iterator = iterator->next) {

		row = g_ptr_array_new();
		g_ptr_array_add(row, g_strdup_printf("%s", ((plugin_stats_aux*)iterator->data)->stat));
		g_ptr_array_add(row, g_strdup_printf("%ld", (long int)((plugin_stats_aux*)iterator->data)->value));

		g_ptr_array_add(rows, row);
	}

	network_mysqld_con_send_resultset(con->client, fields, rows);

	resultset_cleanup(fields, rows, row);
}

/**
 * resultset_stats_all_global:
 * @con: a #network_mysqld_con
 * @list: GSList of #plugin_stats_aux
 *
 * send all the gathered statistics(mysql-proxy and loaded plugins) to the client
 */
static void resultset_stats_all_global(network_mysqld_con *con, GSList *list) {

	GPtrArray *fields, *rows, *row;
	MYSQL_FIELD *field;

	GSList *iterator = NULL;

	fields = NULL;
	rows = NULL;
	row = NULL;

	fields = network_mysqld_proto_fielddefs_new();

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("statistics");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("plugin");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	field = network_mysqld_proto_fielddef_new();
	field->name = g_strdup("value");
	field->type = FIELD_TYPE_VAR_STRING;
	g_ptr_array_add(fields, field);

	rows = g_ptr_array_new();

	for(iterator = list; iterator; iterator = iterator->next) {

		row = g_ptr_array_new();
		g_ptr_array_add(row, g_strdup_printf("%s", ((plugin_stats_aux*)iterator->data)->stat));
		g_ptr_array_add(row, g_strdup_printf("%s", ((plugin_stats_aux*)iterator->data)->plugin));
		g_ptr_array_add(row, g_strdup_printf("%ld", (long int)((plugin_stats_aux*)iterator->data)->value));

		g_ptr_array_add(rows, row);
	}

	network_mysqld_con_send_resultset(con->client, fields, rows);

	resultset_cleanup(fields, rows, row);
}

/**
 * get_plugin_uptime:
 * @chas: mysql-proxy chassis
 * @plugin_name: plugin name
 *
 * gets the uptime for a specific plugin
 *
 * Returns: a time_t containing the uptime value of #plugin_name
 */
static time_t get_plugin_uptime(chassis G_GNUC_UNUSED *chas, char *plugin_name) {

	time_t act_timestamp, plugin_timestamp;
	guint i, uptime=-1;

	if(!g_strcmp0(plugin_name, "global")) {

		time(&act_timestamp);
		uptime = act_timestamp - chas->init_timestamp;
	}

	else {

		for (i = 0; i < chas->modules->len; i++) {
			chassis_plugin *plugin = chas->modules->pdata[i];

			if(!g_ascii_strcasecmp(plugin->name, plugin_name)) {

				if (plugin->local_stats != NULL && plugin->get_local_stats != NULL) {
					/* GHashTable *stats_hash = NULL;
					stats_hash = plugin->get_stats(plugin->stats); */

					plugin_timestamp = chassis_plugin_stats_uptime_get(plugin->local_stats);

					time(&act_timestamp);
					uptime = act_timestamp - plugin_timestamp;
				}
			}
		}
	}

	return uptime;
}

/**
 * plugin_stats_aux_insert:
 * @stat: statistic string identifier
 * @value: statistic value
 * @plugin: plugin string identifier
 *
 * setter for #plugin_stats_aux
 *
 * Returns: a pointer to a #plugin_stats_aux
 * See: #plugin_stats_aux
 */
plugin_stats_aux *plugin_stats_aux_insert(gchar *stat, long int value, gchar *plugin) {

	plugin_stats_aux *plugin_stats = g_new(plugin_stats_aux, 1);
	plugin_stats->stat 	= stat;
	plugin_stats->value = (void*)value;
	plugin_stats->plugin = plugin;

	return plugin_stats;
}

/*
 * DOC-COMMENT THIS!
 */
GSList *get_uptime_all(chassis G_GNUC_UNUSED *chas) {

	GSList *uptime_stats_list = NULL;
	int i;

	uptime_stats_list = g_slist_append(uptime_stats_list,
		 plugin_stats_aux_insert("uptime", get_plugin_uptime(chas, "global"), "(global)"));

	for (i = 0; i < chas->modules->len; i++) {
		chassis_plugin *plugin = chas->modules->pdata[i];

		if (plugin->local_stats != NULL && plugin->get_local_stats != NULL)
			uptime_stats_list = g_slist_append(uptime_stats_list,
				 plugin_stats_aux_insert("uptime", get_plugin_uptime(chas, plugin->name), plugin->name));
	}

	return uptime_stats_list;
}

/*
 * DOC-COMMENT THIS!
 */
GSList *get_latency_all(void) {

	GSList *latency_stats_list = NULL;

	long read_query_ts_min = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(read_query_ts_min);
	long read_query_ts_max = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(read_query_ts_max);
	long read_query_ts_avg = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(read_query_ts_avg);

	latency_stats_list = g_slist_append(latency_stats_list, plugin_stats_aux_insert("minimum query latency (microseconds)",
	read_query_ts_min, "proxy"));

	latency_stats_list = g_slist_append(latency_stats_list, plugin_stats_aux_insert("maximum query latency (microseconds)",
	read_query_ts_max, "proxy"));

	latency_stats_list = g_slist_append(latency_stats_list, plugin_stats_aux_insert("average query latency (microseconds)",
	read_query_ts_avg, "proxy"));

	return latency_stats_list;
}

/*
 * DOC-COMMENT THIS!
 */
GSList *get_connections_all(void) {

	GSList *connections_stats_list = NULL;

	guint connected_clients = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(connected_clients);
	guint connected_clients_max = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(connected_clients_max);

	connections_stats_list = g_slist_append(connections_stats_list, plugin_stats_aux_insert("current connected clients",
	connected_clients, "proxy"));

	connections_stats_list = g_slist_append(connections_stats_list, plugin_stats_aux_insert("maximum connected clients",
	connected_clients_max, "proxy"));

	return connections_stats_list;	
}

/**
 * admin_stats_handler:
 * @chas: mysql-proxy chassis
 * @con: a #network_mysqld_con
 * @ret: a #admin_plugin_stats_query_t
 *
 * handler for all the possible states of #admin_plugin_stats_query_t
 *
 * Returns: a #network_mysqld_lua_stmt_ret
 */
network_mysqld_lua_stmt_ret admin_stats_handler(chassis G_GNUC_UNUSED *chas, network_mysqld_con *con, admin_plugin_stats_query_t ret) {

	switch(ret) {

		case STATS_QUERY_INCOMPLETE: {

			network_mysqld_con_send_error(con->client, C("use 'SHOW help' to see the supported commands"));

			con->state = CON_STATE_SEND_ERROR;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_HELP: {

			resultset_helper(con);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_STATS_AVAILABLE: {

			resultset_stats_available(con);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_STATS_BACKENDS: {

			resultset_stats_backends(con);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_ALL_ADMIN: {

			GSList *stats_list = NULL;

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("uptime", get_plugin_uptime(chas, "admin"), NULL));

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("another admin stats", 12345, NULL));

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("and another admin stat", 54321, NULL));

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("and one more admin stat", 11111, NULL));

			resultset_stats_all_plugin(con, stats_list);

			g_slist_free(stats_list);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_ALL_PROXY: {

			GSList *stats_list = NULL;

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("uptime", get_plugin_uptime(chas, "proxy"), NULL));

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("another proxy stat", 12345, NULL));

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("and another proxy stat", 54321, NULL));

			stats_list = g_slist_append(stats_list, plugin_stats_aux_insert("and one more proxy stat", 11111, NULL));

			resultset_stats_all_plugin(con, stats_list);

			g_slist_free(stats_list);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_ALL_GLOBAL: {

			GSList *stats_list = NULL;

			/* get all the uptime stats and add them to the stats list */
			stats_list = get_uptime_all(chas);

			/* get all the latency stats and add them to the stats list */
			stats_list = g_slist_concat(stats_list, get_latency_all());

			/* get all the connections stats and add them to the stats list */
			stats_list = g_slist_concat(stats_list, get_connections_all());

			resultset_stats_all_global(con, stats_list);

			g_slist_free(stats_list);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_UPTIME_ADMIN: {

			resultset_send("uptime admin plugin (seconds)", get_plugin_uptime(chas, "admin"), con);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_UPTIME_PROXY: {

			resultset_send("uptime proxy plugin (seconds)", get_plugin_uptime(chas, "proxy"), con);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_UPTIME_ALL: {

			GSList *stats_list = NULL;

			/* get all the uptime stats and add to the stats list */
			stats_list = get_uptime_all(chas);

			resultset_stats_all_global(con, stats_list);

			g_slist_free(stats_list);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_CONNECTIONS_ALL: {

			GSList *stats_list = NULL;

			stats_list = get_connections_all();

			resultset_stats_all_global(con, stats_list);

			g_slist_free(stats_list);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_CONNECTIONS_CURRENT: {

			guint connected_clients = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(connected_clients);
			resultset_send("current connected clients", connected_clients, con);

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_CONNECTIONS_MAX: {

			guint connected_clients_max = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(connected_clients_max);
			resultset_send("maximum connected clients", connected_clients_max, con);

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_LATENCY_ALL: {

			GSList *stats_list = NULL;

			stats_list = get_latency_all();

			resultset_stats_all_global(con, stats_list);

			g_slist_free(stats_list);

			con->state = CON_STATE_SEND_QUERY_RESULT;

			return PROXY_SEND_RESULT;
				
		}

		case STATS_QUERY_LATENCY_AVG: {

			long read_query_ts_avg = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(read_query_ts_avg);
			resultset_send("average query latency (microseconds)", read_query_ts_avg, con);

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_LATENCY_MIN: {

			long read_query_ts_min = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(read_query_ts_min);
			resultset_send("minimum query latency (microseconds)", read_query_ts_min, con);

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_LATENCY_MAX: {

			long read_query_ts_max = CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(read_query_ts_max);
			resultset_send("maximum query latency (microseconds)", read_query_ts_max, con);

			return PROXY_SEND_RESULT;
		}

		case STATS_QUERY_SUCCESS:
			/* complete */
			/* network_mysqld_con_send_error(con->client, C("ERROR: X")); */
			return PROXY_NO_DECISION;
		/*
		case STATS_QUERY_ERROR:
			g_error("yyerror()");
			break;
			*/
	}

	return 0;
}
