/* $%BEGINLICENSE%$
 Copyright (c) 2012, Oracle and/or its affiliates. All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; version 2 of the
 License.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 02110-1301  USA

 $%ENDLICENSE%$ */

%{
#include <glib.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "admin-sql-parser.h"
#include "admin-plugin-stats.h"

%}

%define api.pure
%lex-param { sql_tokenizer_t * tokenizer}
%parse-param { sql_tokenizer_t * tokenizer}

/* see sql-tokenizer.h ... and 208 is TK_SQL_SHOW, 9 is TK_LITERAL */
%token SHOW 208
%token LITERAL 9

%%

exp:    SHOW LITERAL {
						/* TODO: add regex for any other literal than 'help' */
						if(!g_strcmp0($2->text->str, "PROXY") ||
							!g_strcmp0($2->text->str, "ADMIN")) {
							return STATS_QUERY_INCOMPLETE;
						}

						else if(!g_strcmp0($2->text->str, "STATUS"))
							return STATS_QUERY_ALL_GLOBAL;

						else if(!g_strcmp0($2->text->str, "UPTIME"))
							return STATS_QUERY_UPTIME_ALL;

						else if(!g_strcmp0($2->text->str, "CONNECTIONS"))
							return STATS_QUERY_CONNECTIONS_ALL;

						else if(!g_strcmp0($2->text->str, "LATENCY"))
							return STATS_QUERY_LATENCY_ALL;

						else if(!g_strcmp0($2->text->str, "HELP"))
							return STATS_QUERY_HELP;
					}
		| SHOW LITERAL LITERAL {
									if(!g_strcmp0($3->text->str, "STATS")) {

										if(!g_strcmp0($2->text->str, "AVAILABLE"))
											return STATS_QUERY_STATS_AVAILABLE;

										else if(!g_strcmp0($2->text->str, "BACKENDS"))
											return STATS_QUERY_STATS_BACKENDS;
									}

									else if(!g_strcmp0($3->text->str, "STATUS")) {

										if(!g_strcmp0($2->text->str, "ADMIN"))
											return STATS_QUERY_ALL_ADMIN;

										else if(!g_strcmp0($2->text->str, "PROXY"))
											return STATS_QUERY_ALL_PROXY;
									}

									else if(!g_strcmp0($3->text->str, "UPTIME")) {

										if(!g_strcmp0($2->text->str, "ADMIN"))
											return STATS_QUERY_UPTIME_ADMIN;

										else if(!g_strcmp0($2->text->str, "PROXY"))
											return STATS_QUERY_UPTIME_PROXY;
									}

									else if(!g_strcmp0($3->text->str, "CONNECTIONS")) {

										if(!g_strcmp0($2->text->str, "CURRENT"))
											return STATS_QUERY_CONNECTIONS_CURRENT;

										else if(!g_strcmp0($2->text->str, "MAX"))
											return STATS_QUERY_CONNECTIONS_MAX;
									}

									else if(!g_strcmp0($3->text->str, "LATENCY")) {

										if(!g_strcmp0($2->text->str, "AVG"))
											return STATS_QUERY_LATENCY_AVG;

										else if(!g_strcmp0($2->text->str, "MIN"))
											return STATS_QUERY_LATENCY_MIN;

										else if(!g_strcmp0($2->text->str, "MAX"))
											return STATS_QUERY_LATENCY_MAX;
									}
								}
;

%%

/**
 * sql_tokenizer_new:
 * @str: string to tokenize
 * @str_len: string length
 *
 * hack the interface of sql_tokenizer() to yylex()'s interface
 *
 * Returns: a pointer to #sql_tokenizer_t
 */
sql_tokenizer_t *
sql_tokenizer_new(const char *str, gsize str_len) {
	sql_tokenizer_t *tokenizer = g_slice_new(sql_tokenizer_t);

	tokenizer->tokens = g_ptr_array_new();
	sql_tokenizer(tokenizer->tokens, str, str_len);

	tokenizer->current_token = 0;

	return tokenizer;
}

/**
 * sql_tokenizer_new:
 * @lvalp: #YYSTYPE
 * @tokenizer: a #sql_tokenizer_t
 *
 * yylex for yyparse()
 *
 * Returns: a int containing the id of the current token
 */
int
yylex(YYSTYPE *lvalp, sql_tokenizer_t *tokenizer) {
	sql_token *current_token;

	if (tokenizer->current_token >= tokenizer->tokens->len) {
		return 0;
	}

	current_token = tokenizer->tokens->pdata[tokenizer->current_token];

	tokenizer->current_token++;

	*lvalp = current_token;

	/*
	g_print("returning %d (%s)\n",
		current_token->token_id,
		sql_token_get_name(current_token->token_id, NULL));
	*/

	return current_token->token_id;
}

/**
 * yyerror:
 * @tokenizer: #sql_tokenizer_t
 * @errmsg: constant string containing the error message
 *
 * yyerror for yyparse()
 */
void
yyerror(sql_tokenizer_t *tokenizer, char const *errmsg) {
	/* printf("error: %s\n", errmsg); */
}

#if 0
int main() {
	sql_tokenizer_t *tokenizer = sql_tokenizer_new("SHOW ADMIN UPTIME\n", 18);

	printf("SHOW = %d\n", TK_SQL_SHOW);
	printf("literal = %d\n", TK_LITERAL);

	return yyparse(tokenizer);
}
#endif
