/* $%BEGINLICENSE%$
 Copyright (c) 2012, Oracle and/or its affiliates. All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; version 2 of the
 License.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 02110-1301  USA

 $%ENDLICENSE%$ */


#ifndef _CHASSIS_PLUGIN_STATS_H_
#define _CHASSIS_PLUGIN_STATS_H_

#include <glib.h>
#include "chassis-exports.h"
#include "network-backend.h"

/**
 * ts_read_query_stats:
 * @get_timestamp: timestamp of the obtainance of the #ts_diff
 * @ts_diff: difference between #ts_read_query_result_last and #ts_read_query_result_first
 *
 * auxiliary struct to store #get_timestamp and #ts_diff and keep on the list #ts_read_query_stats_list
 */
typedef struct {
	time_t get_timestamp;
	guint64 ts_diff;
} ts_read_query_stats;

extern network_backends_t *backends_stats;

/**
 * chassis_plugin_local_stats_t:
 * @plugin_init_timestamp: time_t of the initial timestamp of the plugin
 * @ts_read_query_stats_list: GList of #ts_read_query_stats
 * @read_query_ts_avg: average of query latency
 *
 * structure to store the local plugin statistics
 */
typedef struct chassis_plugin_local_stats {
	volatile time_t plugin_init_timestamp;
	volatile guint read_query_ts_avg;
	volatile guint read_query_ts_min;
	volatile guint read_query_ts_max;
	volatile guint connected_clients;
	volatile guint connected_clients_max;
} chassis_plugin_local_stats_t;

extern chassis_plugin_local_stats_t *plugin_global_stats;
extern GList *ts_read_query_stats_list;

CHASSIS_API chassis_plugin_local_stats_t * chassis_plugin_stats_new(chassis_plugin_local_stats_t *stats);
CHASSIS_API void chassis_plugin_stats_free(chassis_plugin_local_stats_t *stats);

CHASSIS_API GHashTable* chassis_plugin_stats_get(chassis_plugin_local_stats_t *user_data);

CHASSIS_API void chassis_plugin_stats_uptime_set(chassis_plugin_local_stats_t *user_data, time_t timestamp);
CHASSIS_API time_t chassis_plugin_stats_uptime_get(chassis_plugin_local_stats_t *user_data);
CHASSIS_API time_t *chassis_plugin_stats_uptime_get_pointer(chassis_plugin_local_stats_t *user_data);

CHASSIS_API void chassis_plugin_stats_read_query_stats_set(time_t timestamp, guint ts_diff);

CHASSIS_API guint chassis_plugin_stats_read_query_stats_get_avg(void);
CHASSIS_API guint chassis_plugin_stats_read_query_stats_get_min(void);
CHASSIS_API guint chassis_plugin_stats_read_query_stats_get_max(void);

#define CHASSIS_PLUGIN_STATS_ALLOC_INC_NAME(stats, name) ((stats != NULL) ? g_atomic_int_inc(&(stats->name ## _alloc)) : (void)0)
#define CHASSIS_PLUGIN_STATS_FREE_INC_NAME(stats, name) ((stats != NULL) ? g_atomic_int_inc(&(stats->name ## _free)) : (void)0)
#define CHASSIS_PLUGIN_STATS_ADD_NAME(stats, name, addme) ((stats != NULL) ? g_atomic_int_add(&(stats->name), addme) : (void)0)
#define CHASSIS_PLUGIN_STATS_GET_NAME(stats, name) ((stats != NULL) ? g_atomic_int_get(&(stats->name)) : 0)
#define CHASSIS_PLUGIN_STATS_SET_NAME(stats, name, setme) ((stats != NULL) ? g_atomic_int_set(&(stats->name), setme) : (void)0)

#define CHASSIS_PLUGIN_STATS_SET_NAME_GLOBAL(name, setme) ((plugin_global_stats != NULL) ? g_atomic_int_set(&(plugin_global_stats->name), setme) : (void)0);
#define CHASSIS_PLUGIN_STATS_GET_NAME_GLOBAL(name) ((plugin_global_stats != NULL) ? g_atomic_int_get(&(plugin_global_stats->name)) : 0);

#endif /* _CHASSIS_PLUGIN_STATS_H_ */